"""
Implements Supply-Chain Firewall's `configure` subcommand.
"""

from argparse import Namespace

import scfw.configure.dd_agent as dd_agent
import scfw.configure.env as env
import scfw.configure.interactive as interactive
from scfw.configure.interactive import GREETING


def run_configure(args: Namespace) -> int:
    """
    Configure the environment for use with the supply-chain firewall.

    Args:
        args: A `Namespace` containing the parsed `configure` subcommand command line.

    Returns:
        An integer status code indicating normal exit.
    """
    if args.remove:
        # These options result in the firewall's configuration block being removed
        env.update_config_files({
            "alias_npm": False,
            "alias_pip": False,
            "alias_poetry": False,
            "dd_agent_port": None,
            "dd_api_key": None,
            "dd_log_level": None,
            "scfw_home": None,
        })
        dd_agent.remove_agent_logging()
        print(
            "All Supply-Chain Firewall-managed configuration has been removed from your environment."
            "\n\nPost-removal tasks:"
            "\n* Update your current shell environment by sourcing from your .bashrc/.zshrc file."
            "\n* If you had previously configured Datadog Agent log forwarding, restart the Agent."
        )
        return 0

    # The CLI parser guarantees that all of these arguments are present
    is_interactive = not any({
        args.alias_npm,
        args.alias_pip,
        args.alias_poetry,
        args.dd_agent_port,
        args.dd_api_key,
        args.dd_log_level,
        args.scfw_home,
    })

    if is_interactive:
        print(GREETING)
        answers = interactive.get_answers()
    else:
        answers = vars(args)

    if not answers:
        return 0

    env.update_config_files(answers)

    if (port := answers.get("dd_agent_port")):
        dd_agent.configure_agent_logging(port)

    if is_interactive:
        print(interactive.get_farewell(answers))

    return 0



"""
A representation of software packages in supported ecosystems.
"""

from dataclasses import dataclass

from scfw.ecosystem import ECOSYSTEM


@dataclass(eq=True, frozen=True)
class Package:
    """
    Specifies a software package in a supported ecosystem.

    Attributes:
        ecosystem: The package's ecosystem.
        name: The package's name.
        version: The package's version string.
    """
    ecosystem: ECOSYSTEM
    name: str
    version: str

    def __str__(self) -> str:
        """
        Represent a `Package` as a string using ecosystem-specific formatting.

        Returns:
            A `str` with ecosystem-specific formatting describing the `Package` name and version.

            `npm` packages: `"{name}@{version}"`.
            `PyPI` packages: `"{name}-{version}"`
        """
        match self.ecosystem:
            case ECOSYSTEM.Npm:
                return f"{self.name}@{self.version}"
            case ECOSYSTEM.PyPI:
                return f"{self.name}-{self.version}"



"""
Utilities for downloading and caching the malicious package dataset manifests.
"""

import glob
import json
import logging
import os
from pathlib import Path
import re
from typing import Optional, TypeAlias

import requests

from scfw.ecosystem import ECOSYSTEM

_log = logging.getLogger(__name__)

_DD_DATASET_SAMPLES_URL = "https://raw.githubusercontent.com/DataDog/malicious-software-packages-dataset/main/samples"

Manifest: TypeAlias = dict[str, list[str]]
"""
A malicious packages dataset manifest mapping package names to affected versions.
"""


def download_manifest(ecosystem: ECOSYSTEM) -> tuple[Optional[str], Manifest]:
    """
    Download the dataset manifest for the given `ecosystem`.
    """
    request = requests.get(_manifest_url(ecosystem), timeout=5)
    request.raise_for_status()

    return (_extract_etag_header(request.headers.get("ETag", "")), request.json())


def get_latest_manifest(cache_dir: Path, ecosystem: ECOSYSTEM) -> Manifest:
    """
    Return the dataset manifest for the given `ecosystem`, either from the given `cache_dir`
    or from the remote dataset, and update the local cache accordingly.

    The names of the cached manifest files adhere to the format `<ecosystem><tag>.json`,
    where `tag` is the (possibly empty) value of the `"ETag"` header obtained on the
    most recent refresh of the manifest files from the dataset on GitHub.
    """
    update_cache = True
    cached_manifest_file = None
    latest_etag, latest_manifest = None, None

    try:
        if (
            (manifest_files := glob.glob(str(cache_dir / f"{ecosystem}*.json")))
            and (cached_manifest_file := Path(manifest_files[0]))
            and (last_etag := _extract_etag_file_name(ecosystem, cached_manifest_file.name))
        ):
            latest_etag, latest_manifest = _update_manifest(ecosystem, last_etag)

            if latest_etag == last_etag:
                _log.debug(f"Cached malicious {ecosystem} packages dataset is up-to-date")
                update_cache = False
                with open(cached_manifest_file) as f:
                    latest_manifest = json.load(f)
        else:
            latest_etag, latest_manifest = download_manifest(ecosystem)

    except Exception:
        _log.warning(f"Failed to obtain malicious {ecosystem} packages metadata or dataset from GitHub")

        if cached_manifest_file:
            _log.warning(f"Using cached copy of malicious {ecosystem} packages dataset after GitHub failure")
            update_cache = False
            with open(cached_manifest_file) as f:
                latest_manifest = json.load(f)

    if not latest_manifest:
        raise RuntimeError(f"Failed to obtain malicious {ecosystem} packages dataset from GitHub or cache")

    if update_cache:
        try:
            if cached_manifest_file:
                _log.debug(f"Removing outdated malicious {ecosystem} packages dataset cache")
                os.remove(cached_manifest_file)

            _log.debug(f"Updating malicious {ecosystem} packages dataset cache")
            cached_manifest_file = cache_dir / f"{ecosystem}{latest_etag if latest_etag else ''}.json"
            if not cached_manifest_file.parent.is_dir():
                cached_manifest_file.parent.mkdir(parents=True, exist_ok=True)
            with open(cached_manifest_file, 'w') as f:
                json.dump(latest_manifest, f)

        except Exception as e:
            _log.warning(f"Failed to update cached copy of malicious {ecosystem} packages dataset: {e}")

    return latest_manifest


def _update_manifest(ecosystem: ECOSYSTEM, etag: str) -> tuple[Optional[str], Optional[Manifest]]:
    """
    Get the latest dataset manifest for the given `ecosystem` relative to the given `etag`.
    """
    request = requests.get(_manifest_url(ecosystem), headers={"If-None-Match": f'W/"{etag}"'}, timeout=5)
    request.raise_for_status()

    if request.status_code == requests.codes.NOT_MODIFIED:
        return (etag, None)

    return (_extract_etag_header(request.headers.get("ETag", "")), request.json())


def _extract_etag_header(s: str) -> Optional[str]:
    """
    Extract an ETag from the given header string.
    """
    match = re.search('W/"(.*)"', s)
    return match.group(1) if match else None


def _extract_etag_file_name(ecosystem: ECOSYSTEM, s: str) -> Optional[str]:
    """
    Extract an ETag from the file name of the given `ecosystem` manifest cache.
    """
    match = re.search(f"{ecosystem}(.*).json", s)
    return match.group(1) if match else None


def _manifest_url(ecosystem: ECOSYSTEM) -> str:
    """
    Construct the GitHub download URL for the given `ecosystem` manifest file.
    """
    return f"{_DD_DATASET_SAMPLES_URL}/{str(ecosystem).lower()}/manifest.json"



"""
Provides a `FirewallLogger` class for sending logs to Datadog.
"""

import getpass
import json
import logging
import os
import socket

import dotenv

import scfw
from scfw.constants import DD_ENV, DD_LOG_LEVEL_VAR, DD_SERVICE, DD_SOURCE
from scfw.ecosystem import ECOSYSTEM
from scfw.logger import FirewallAction, FirewallLogger
from scfw.package import Package
from scfw.report import VerificationReport
from scfw.verifier import FindingSeverity

_log = logging.getLogger(__name__)

_DD_LOG_LEVEL_DEFAULT = FirewallAction.BLOCK

# The `created` and `msg` attributes are provided by `logging.LogRecord`
_AUDIT_ATTRIBUTES = {
    "created",
    "ecosystem",
    "executable",
    "msg",
    "package_manager",
    "reports",
}
_FIREWALL_ACTION_ATTRIBUTES = {
    "action",
    "created",
    "ecosystem",
    "executable",
    "msg",
    "package_manager",
    "targets",
    "verified",
    "warned",
}


dotenv.load_dotenv()


class DDLogFormatter(logging.Formatter):
    """
    A custom JSON formatter for firewall logs.
    """
    def format(self, record) -> str:
        """
        Format a log record as a JSON string.

        Args:
            record: The log record to be formatted.
        """
        log_record = {
            "source": DD_SOURCE,
            "service": DD_SERVICE,
            "version": scfw.__version__,
            "env": os.getenv("DD_ENV", DD_ENV),
            "hostname": socket.gethostname(),
        }

        try:
            log_record["username"] = getpass.getuser()
        except Exception as e:
            _log.warning(f"Failed to query username while formatting log: {e}")

        for key in _AUDIT_ATTRIBUTES | _FIREWALL_ACTION_ATTRIBUTES:
            try:
                log_record[key] = record.__dict__[key]
            except KeyError:
                pass

        return json.dumps(log_record) + '\n'


class DDLogger(FirewallLogger):
    """
    An implementation of `FirewallLogger` for sending logs to Datadog.
    """
    def __init__(self, logger: logging.Logger):
        """
        Initialize a new `DDLogger`.

        Args:
            logger: A configured log handle to which logs will be written.
        """
        self._logger = logger
        self._level = _DD_LOG_LEVEL_DEFAULT

        try:
            if (dd_log_level := os.getenv(DD_LOG_LEVEL_VAR)) is not None:
                self._level = FirewallAction.from_string(dd_log_level)
        except ValueError:
            _log.warning(f"Undefined or invalid Datadog log level: using default level {_DD_LOG_LEVEL_DEFAULT}")

    def log_firewall_action(
        self,
        ecosystem: ECOSYSTEM,
        package_manager: str,
        executable: str,
        command: list[str],
        targets: list[Package],
        action: FirewallAction,
        verified: bool,
        warned: bool,
    ):
        """
        Log the data and action taken in a completed run of Supply-Chain Firewall.

        Args:
            ecosystem: The ecosystem of the inspected package manager command.
            package_manager: The command-line name of the package manager.
            executable: The executable used to execute the inspected package manager command.
            command: The package manager command line provided to the firewall.
            targets: The installation targets relevant to firewall's action.
            action: The action taken by the firewall.
            verified: Indicates whether verification was performed in taking the specified `action`.
            warned: Indicates whether the user was warned about findings and prompted for approval.
        """
        if not self._level or action < self._level:
            return

        self._logger.info(
            f"Command '{' '.join(command)}' was {str(action).lower()}ed",
            extra={
                "ecosystem": str(ecosystem),
                "package_manager": package_manager,
                "executable": executable,
                "targets": list(map(str, targets)),
                "action": str(action),
                "verified": verified,
                "warned": warned,
            }
        )

    def log_audit(
        self,
        ecosystem: ECOSYSTEM,
        package_manager: str,
        executable: str,
        reports: dict[FindingSeverity, VerificationReport],
    ):
        """
        Log the results of an audit for the given ecosystem and package manager.

        Args:
            ecosystem: The ecosystem of the audited packages.
            package_manager: The package manager that manages the audited packages.
            executable: The package manager executable used to enumerate audited packages.
            reports: The severity-ranked reports resulting from auditing the installed packages.
        """
        self._logger.info(
            f"Successfully audited {ecosystem} packages managed by {package_manager}",
            extra={
                "ecosystem": str(ecosystem),
                "package_manager": package_manager,
                "executable": executable,
                "reports": {
                    str(severity): list(map(str, report.packages())) for severity, report in reports.items()
                },
            }
        )



"""
Provides a representation of OSV advisories for use in `OsvVerifier`.
"""

from dataclasses import dataclass
from enum import Enum
from typing import Optional
from typing_extensions import Self

from cvss import CVSS2, CVSS3, CVSS4  # type: ignore


class Severity(Enum):
    """
    Represents the possible severities that an OSV advisory can have.
    """
    Non = 0
    Low = 1
    Medium = 2
    High = 3
    Critical = 4

    def __lt__(self, other: Self) -> bool:
        """
        Compare two `Severity` instances.

        Args:
            self: The `Severity` to be compared on the left-hand side
            other: The `Severity` to be compared on the right-hand side

        Returns:
            A `bool` indicating whether `<` holds between the two given `Severity`.

        Raises:
            TypeError: The other argument given was not a `Severity`.
        """
        if self.__class__ is not other.__class__:
            raise TypeError(
                f"'<' not supported between instances of '{self.__class__}' and '{other.__class__}'"
            )

        return self.value < other.value

    def __str__(self) -> str:
        """
        Format a `Severity` for printing.

        Returns:
            A `str` representing the given `Severity` suitable for printing.
        """
        return "None" if self.name == "Non" else self.name

    @classmethod
    def from_string(cls, s: str) -> Self:
        """
        Convert a string into a `Severity`.

        Args:
            s: The `str` to be converted.

        Returns:
            The `Severity` referred to by the given string.

        Raises:
            ValueError: The given string does not refer to a valid `Severity`.
        """
        mappings = {f"{severity}".lower(): severity for severity in cls}
        if (severity := mappings.get(s.lower())):
            return severity
        raise ValueError(f"Invalid severity '{s}'")


class OsvSeverityType(str, Enum):
    """
    The various severity score types defined in the OSV standard.
    """
    CVSS_V2 = "CVSS_V2"
    CVSS_V3 = "CVSS_V3"
    CVSS_V4 = "CVSS_V4"
    Ubuntu = "Ubuntu"


@dataclass(eq=True, frozen=True)
class OsvSeverityScore:
    """
    A typed severity score used in assigning severities to OSV advisories.
    """
    type: OsvSeverityType
    score: str

    @classmethod
    def from_json(cls, osv_json: dict) -> Self:
        """
        Convert a JSON-formatted OSV advisory into an `OsvSeverityScore`.

        Args:
            osv_json: The JSON-formatted OSV severity score to be converted.

        Returns:
            An `OsvSeverityScore` derived from the content of the given JSON.

        Raises:
            ValueError: The severity score was malformed or missing required information.
        """
        type = osv_json.get("type")
        score = osv_json.get("score")
        if type and score:
            return cls(type=OsvSeverityType(type), score=score)
        raise ValueError("Encountered malformed OSV severity score")

    def severity(self) -> Severity:
        """
        Return the `Severity` of the given `OsvSeverityScore`.

        Returns:
            The computed `Severity` of the given `OsvSeverityScore`.
        """
        match self.type:
            case OsvSeverityType.CVSS_V2:
                severity_str = CVSS2(self.score).severities()[0]
            case OsvSeverityType.CVSS_V3:
                severity_str = CVSS3(self.score).severities()[0]
            case OsvSeverityType.CVSS_V4:
                severity_str = CVSS4(self.score).severity
            case OsvSeverityType.Ubuntu:
                severity_str = "None" if self.score == "Negligible" else self.score

        return Severity.from_string(severity_str)


@dataclass(eq=True, frozen=True)
class OsvAdvisory:
    """
    A representation of an OSV advisory containing only the fields relevant to
    package verification.
    """
    id: str
    severity: Optional[Severity]

    @classmethod
    def compare_severities(cls, lhs: Self, rhs: Self) -> int:
        """
        Compare two `OsvAdvisory` instances on the basis of their severities such that
        advisories with no severities are sorted lower than those with severities.

        Args:
            self: The `OsvAdvisory` to be compared on the left-hand side
            other: The `OsvAdvisory` to be compared on the right-hand side

        Returns:
            An `int` indicating whether the first `OsvAdvisory` is less than, equal to
            or greater than the second one.

        Raises:
            TypeError: One of the given arguments is not an `OsvAdvisory`.
        """
        if not (isinstance(lhs, cls) and isinstance(rhs, cls)):
            raise TypeError("Received incompatible argument types while comparing OSV severities")

        # A match statement would be more natural here but mypy is not up to it
        if lhs.severity == rhs.severity:
            result = 0
        elif lhs.severity is None:
            result = -1
        elif rhs.severity is None:
            result = 1
        elif lhs.severity < rhs.severity:
            result = -1
        elif rhs.severity < lhs.severity:
            result = 1

        return result

    @classmethod
    def from_json(cls, osv_json: dict) -> Self:
        """
        Convert a JSON-formatted OSV advisory into an `OsvAdvisory`.

        Args:
            osv_json: The JSON-formatted OSV advisory to be converted.

        Returns:
            An `OsvAdvisory` derived from the content of the given JSON.

        Raises:
            ValueError: The advisory was malformed or missing required information.
        """
        if (id := osv_json.get("id")):
            scores = list(map(OsvSeverityScore.from_json, osv_json.get("severity", [])))
            return cls(
                id=id,
                severity=max(map(lambda score: score.severity(), scores)) if scores else None
            )
        raise ValueError("Encountered OSV advisory with missing ID field")



"""
Implements the supply-chain firewall's core `run` subcommand.
"""

from argparse import Namespace
import inquirer  # type: ignore
import logging
import os
import sys

from scfw.constants import ON_WARNING_VAR
from scfw.logger import FirewallAction
from scfw.loggers import FirewallLoggers
from scfw.package_manager import UnsupportedVersionError
import scfw.package_managers as package_managers
from scfw.verifier import FindingSeverity
from scfw.verifiers import FirewallVerifiers

_log = logging.getLogger(__name__)


def run_firewall(args: Namespace) -> int:
    """
    Run a package manager command through the supply-chain firewall.

    Args:
        args:
            A `Namespace` parsed from a `run` subcommand command line containing a
            command to run through the firewall.

    Returns:
        An integer status code indicating normal or error exit.
    """
    package_manager = None
    critical_report, warning_report = None, None

    loggers = FirewallLoggers()
    _log.info(f"Command: '{' '.join(args.command)}'")

    try:
        package_manager = package_managers.get_package_manager(args.package_manager, executable=args.executable)

        targets = package_manager.resolve_install_targets(args.command)
        _log.info(f"Command would install: [{', '.join(map(str, targets))}]")

        if targets:
            verifiers = FirewallVerifiers(package_manager.ecosystem())
            _log.info(f"Using package verifiers: [{', '.join(verifiers.names())}]")

            reports = verifiers.verify_packages(targets)
            critical_report = reports.get(FindingSeverity.CRITICAL)
            warning_report = reports.get(FindingSeverity.WARNING)

            if not args.dry_run and critical_report:
                loggers.log_firewall_action(
                    package_manager.ecosystem(),
                    package_manager.name(),
                    package_manager.executable(),
                    args.command,
                    list(critical_report.packages()),
                    action=FirewallAction.BLOCK,
                    verified=True,
                    warned=False,
                )
                print(critical_report)
                print("\nThe installation request was blocked. No changes have been made.")
                return 1 if args.error_on_block else 0

            if not args.dry_run and warning_report:
                print(warning_report)
                if _get_warning_action(args.allow_on_warning, args.block_on_warning) == FirewallAction.BLOCK:
                    loggers.log_firewall_action(
                        package_manager.ecosystem(),
                        package_manager.name(),
                        package_manager.executable(),
                        args.command,
                        list(warning_report.packages()),
                        action=FirewallAction.BLOCK,
                        verified=True,
                        warned=True,
                    )
                    print("The installation request was aborted. No changes have been made.")
                    return 1 if args.error_on_block else 0

        if args.dry_run:
            _log.info("Firewall dry-run mode enabled: command will not be run")
            if critical_report:
                print(critical_report)
            elif warning_report:
                print(warning_report)
            print("Dry-run: exiting without running command.")
            return 1 if (critical_report or warning_report) else 0

        loggers.log_firewall_action(
            package_manager.ecosystem(),
            package_manager.name(),
            package_manager.executable(),
            args.command,
            targets,
            action=FirewallAction.ALLOW,
            verified=True,
            warned=True if warning_report else False,
        )
        return package_manager.run_command(args.command)

    except UnsupportedVersionError as e:
        if not args.allow_unsupported:
            _log.error(e)
            _log.error(
                "Upgrade to a supported version or rerun with --allow-unsupported to bypass verification (use caution)"
            )
            return 0

        _log.info(f"Unsupported package manager version: {e}")
        _log.info(f"Unsupported versions allowed: running command '{' '.join(args.command)}' without verification")

        if not package_manager:
            raise RuntimeError("Failed to initialize package manager handle: cannot run command")

        loggers.log_firewall_action(
            package_manager.ecosystem(),
            package_manager.name(),
            package_manager.executable(),
            args.command,
            targets=[],
            action=FirewallAction.ALLOW,
            verified=False,
            warned=False,
        )
        return package_manager.run_command(args.command)


def _get_warning_action(cli_allow_choice: bool, cli_block_choice: bool) -> FirewallAction:
    """
    Return the `FirewallAction` that should be taken for `WARNING`-level findings.

    Args:
        cli_allow_choice:
            A `bool` indicating whether the user selected `--allow-on-warning` on the command-line.
        cli_block_choice:
            A `bool` indicating whether the user selected `--block-on-warning` on the command-line.

    Returns:
        The `FirewallAction` that should be taken based on the user's configured choices or, if
        no choice has been made, on the user's runtime (interactive) decision.
    """
    if cli_block_choice:
        return FirewallAction.BLOCK
    if cli_allow_choice:
        return FirewallAction.ALLOW

    if (action := os.getenv(ON_WARNING_VAR)):
        try:
            return FirewallAction.from_string(action)
        except Exception:
            _log.warning(f"Ignoring invalid firewall action {ON_WARNING_VAR}='{action}'")

    if not sys.stdin.isatty():
        _log.warning(
            "Non-interactive terminal and no predefined action for WARNING findings: defaulting to BLOCK"
        )
        return FirewallAction.BLOCK

    user_confirmed = inquirer.confirm("Proceed with installation?", default=False)
    return FirewallAction.ALLOW if user_confirmed else FirewallAction.BLOCK



"""
Provides an interface for client loggers to receive information about a
completed run of the supply-chain firewall.
"""

from abc import (ABCMeta, abstractmethod)
from enum import Enum
from typing_extensions import Self

from scfw.ecosystem import ECOSYSTEM
from scfw.package import Package
from scfw.report import VerificationReport
from scfw.verifier import FindingSeverity


class FirewallAction(Enum):
    """
    The various actions the firewall may take in response to inspecting a
    package manager command.
    """
    ALLOW = 0
    BLOCK = 1

    def __lt__(self, other) -> bool:
        """
        Compare two `FirewallAction` instances on the basis of their underlying numeric values.

        Args:
            self: The `FirewallAction` to be compared on the left-hand side
            other: The `FirewallAction` to be compared on the right-hand side

        Returns:
            A `bool` indicating whether `<` holds between the two given `FirewallAction`.

        Raises:
            TypeError: The other argument given was not a `FirewallAction`.
        """
        if self.__class__ is not other.__class__:
            raise TypeError(
                f"'<' not supported between instances of '{self.__class__}' and '{other.__class__}'"
            )

        return self.value < other.value

    def __str__(self) -> str:
        """
        Format a `FirewallAction` for printing.

        Returns:
            A `str` representing the given `FirewallAction` suitable for printing.
        """
        return self.name

    @classmethod
    def from_string(cls, s: str) -> Self:
        """
        Convert a string into a `FirewallAction`.

        Args:
            s: The `str` to be converted.

        Returns:
            The `FirewallAction` referred to by the given string.

        Raises:
            ValueError: The given string does not refer to a valid `FirewallAction`.
        """
        mappings = {f"{action}".lower(): action for action in cls}
        if (action := mappings.get(s.lower())):
            return action
        raise ValueError(f"Invalid firewall action '{s}'")


class FirewallLogger(metaclass=ABCMeta):
    """
    An interface for passing information about runs of Supply-Chain Firewall to
    client loggers.
    """
    @abstractmethod
    def log_firewall_action(
        self,
        ecosystem: ECOSYSTEM,
        package_manager: str,
        executable: str,
        command: list[str],
        targets: list[Package],
        action: FirewallAction,
        verified: bool,
        warned: bool,
    ):
        """
        Log the data and action taken in a completed run of Supply-Chain Firewall.

        Args:
            ecosystem: The ecosystem of the inspected package manager command.
            package_manager: The command-line name of the package manager.
            executable: The executable used to execute the inspected package manager command.
            command: The package manager command line provided to the firewall.
            targets:
                The installation targets relevant to Supply-Chain Firewall's action:
                  * For `BLOCK` actions, contains the installation targets that caused the block
                  * For `ALLOW` actions, contains all installation targets
            action: The action taken by Supply-Chain Firewall.
            verified:
                Indicates whether Supply-Chain Firewall performed installation target
                verification in deciding to take the specified `action`. Verification is not
                performed **only** under the following conditions:
                  * The package manager is of an unsupported version and the user has passed the
                    command-line option `--allow-unsupported`
            warned:
                Indicates whether the user was warned about findings for any installation
                targets and prompted for approval to proceed with `command`.
        """
        pass

    @abstractmethod
    def log_audit(
        self,
        ecosystem: ECOSYSTEM,
        package_manager: str,
        executable: str,
        reports: dict[FindingSeverity, VerificationReport],
    ):
        """
        Log the results of an audit for the given ecosystem and package manager.

        Args:
            ecosystem: The ecosystem of the audited packages.
            package_manager: The package manager that manages the audited packages.
            executable: The package manager executable used to enumerate audited packages.
            reports:
                The severity-ranked reports resulting from auditing the installed packages.

                These reports contain only those packages for which at least one verifier had
                a finding (at the severity level associated with the entire report).  That is,
                packages with no findings are excluded from the audit results.
        """
        pass



"""
Tests of utilities for writing Supply-Chain Firewall configuration to supported files.
"""

from pathlib import Path
import pytest
from tempfile import NamedTemporaryFile

from scfw.configure.env import _BLOCK_END, _BLOCK_START
import scfw.configure.env as env

ORIGINAL_CONFIG = """\
# Set an environment variable
export MY_ENV_VAR=foo
"""

SCFW_CONFIG_BASE = """\
alias npm="scfw run npm"
alias pip="scfw run pip"
export SCFW_DD_AGENT_LOG_PORT="10365"
export DD_LOG_LEVEL="ALLOW"
"""

SCFW_CONFIG_UPDATED = """\
alias poetry="scfw run poetry"
export SCFW_HOME="~/.scfw"
"""


def enclose(scfw_config: str) -> str:
    """
    Enclose the given `scfw_config` in its block comments.
    """
    return f"{_BLOCK_START}\n{scfw_config}{_BLOCK_END}"


@pytest.mark.parametrize(
        "original_config,scfw_config,updated_config",
        [
            # Initial configuration of an empty file
            (
                "",
                SCFW_CONFIG_BASE,
                f"\n{enclose(SCFW_CONFIG_BASE)}\n",
            ),
            # Initial configuration of a nonempty file
            (
                ORIGINAL_CONFIG,
                SCFW_CONFIG_BASE,
                f"{ORIGINAL_CONFIG}\n{enclose(SCFW_CONFIG_BASE)}\n",
            ),
            # Update configuration when there is no content inside the SCFW block
            (
                enclose(""),
                SCFW_CONFIG_UPDATED,
                enclose(SCFW_CONFIG_UPDATED),
            ),
            # Update configuration in an otherwise empty file with leading and
            # trailing whitespace (as would be added when we configure initially)
            (
                f"\n{enclose(SCFW_CONFIG_BASE)}\n",
                SCFW_CONFIG_UPDATED,
                f"\n{enclose(SCFW_CONFIG_UPDATED)}\n",
            ),
            # Update configuration in an otherwise empty file with no surrounding whitespace
            (
                enclose(SCFW_CONFIG_BASE),
                SCFW_CONFIG_UPDATED,
                enclose(SCFW_CONFIG_UPDATED),
            ),
            # Update configuration at the end of a nonempty file where the SCFW
            # configuration block is separated from surrounding content by whitespace
            (
                f"{ORIGINAL_CONFIG}\n{enclose(SCFW_CONFIG_BASE)}\n",
                SCFW_CONFIG_UPDATED,
                f"{ORIGINAL_CONFIG}\n{enclose(SCFW_CONFIG_UPDATED)}\n",
            ),
            # Update configuration at the end of a nonempty file where the SCFW
            # configuration block is not separated from surrounding content by whitespace
            (
                f"{ORIGINAL_CONFIG}{enclose(SCFW_CONFIG_BASE)}",
                SCFW_CONFIG_UPDATED,
                f"{ORIGINAL_CONFIG}{enclose(SCFW_CONFIG_UPDATED)}",
            ),
            # Update configuration in the middle of a nonempty file where the SCFW
            # configuration block is separated from surrounding content by whitespace
            (
                f"{ORIGINAL_CONFIG}\n{enclose(SCFW_CONFIG_BASE)}\n{ORIGINAL_CONFIG}",
                SCFW_CONFIG_UPDATED,
                f"{ORIGINAL_CONFIG}\n{enclose(SCFW_CONFIG_UPDATED)}\n{ORIGINAL_CONFIG}",
            ),
            # Update configuration in the middle of a nonempty file where the SCFW
            # configuration block is not separated from surrounding content by whitespace
            (
                f"{ORIGINAL_CONFIG}{enclose(SCFW_CONFIG_BASE)}{ORIGINAL_CONFIG}",
                SCFW_CONFIG_UPDATED,
                f"{ORIGINAL_CONFIG}{enclose(SCFW_CONFIG_UPDATED)}{ORIGINAL_CONFIG}",
            ),
            # Remove configuration from an empty file
            (
                "",
                "",
                "",
            ),
            # Remove configuration from a file that contains no configuration
            (
                ORIGINAL_CONFIG,
                "",
                ORIGINAL_CONFIG,
            ),
            # Remove configuration from an otherwise empty file with no leading or
            # trailing whitespace
            (
                enclose(SCFW_CONFIG_BASE),
                "",
                "",
            ),
            # Remove configuration from an otherwise empty file with leading and
            # trailing whitespace (as would be added when we configure initially)
            (
                f"\n{enclose(SCFW_CONFIG_BASE)}\n",
                "",
                "\n\n",
            ),
            # Remove configuration from the end of a nonempty file where the SCFW
            # configuration block is separated from surrounding content by whitespace
            (
                f"{ORIGINAL_CONFIG}\n{enclose(SCFW_CONFIG_BASE)}\n",
                "",
                f"{ORIGINAL_CONFIG}\n\n"
            ),
            # Remove configuration from the end of a nonempty file where the SCFW
            # configuration block is not separated from surrounding content by whitespace
            (
                f"{ORIGINAL_CONFIG}{enclose(SCFW_CONFIG_BASE)}",
                "",
                ORIGINAL_CONFIG,
            ),
            # Remove configuration from the middle of a nonempty file where the SCFW
            # configuation block is separated from surrounding content by whitespace
            (
                f"{ORIGINAL_CONFIG}\n{enclose(SCFW_CONFIG_BASE)}\n{ORIGINAL_CONFIG}",
                "",
                f"{ORIGINAL_CONFIG}\n\n{ORIGINAL_CONFIG}",
            ),
            # Remove configuration from the middle of a nonempty file where the SCFW
            # configuation block is not separated from surrounding content by whitespace
            (
                f"{ORIGINAL_CONFIG}{enclose(SCFW_CONFIG_BASE)}{ORIGINAL_CONFIG}",
                "",
                f"{ORIGINAL_CONFIG}{ORIGINAL_CONFIG}",
            ),
        ]
)
def test_config_file_update(original_config: str, scfw_config: str, updated_config: str):
    """
    Test that an update to configuration file contents has the expected result.
    """
    with NamedTemporaryFile(mode="r+") as f:
        if original_config:
            f.write(original_config)
            f.seek(0)

        env._update_config_file(Path(f.name), scfw_config)

        content = f.read()
        print(f"'{content}'")

        assert content == updated_config



"""
Defines a package verifier for Datadog Security Research's malicious packages dataset.
"""

import os
from pathlib import Path

from scfw.constants import SCFW_HOME_VAR
from scfw.ecosystem import ECOSYSTEM
from scfw.package import Package
from scfw.verifier import FindingSeverity, PackageVerifier
import scfw.verifiers.dd_verifier.dataset as dataset


class DatadogMaliciousPackagesVerifier(PackageVerifier):
    """
    A `PackageVerifier` for Datadog Security Research's malicious packages dataset.
    """
    def __init__(self):
        """
        Initialize a new `DatadogMaliciousPackagesVerifier`.
        """
        self._manifests = {}

        cache_dir = None
        if (home_dir := os.getenv(SCFW_HOME_VAR)):
            cache_dir = Path(home_dir) / "dd_verifier"

        for ecosystem in self.supported_ecosystems():
            if cache_dir:
                self._manifests[ecosystem] = dataset.get_latest_manifest(cache_dir, ecosystem)
            else:
                _, self._manifests[ecosystem] = dataset.download_manifest(ecosystem)

    @classmethod
    def name(cls) -> str:
        """
        Return the `DatadogMaliciousPackagesVerifier` name string.

        Returns:
            The class' constant name string: `"DatadogMaliciousPackagesVerifier"`.
        """
        return "DatadogMaliciousPackagesVerifier"

    @classmethod
    def supported_ecosystems(cls) -> set[ECOSYSTEM]:
        """
        Return the set of package ecosystems supported by `DatadogMaliciousPackagesVerifier`.

        Returns:
            The class' constant set of supported ecosystems: `{ECOSYSTEM.Npm, ECOSYSTEM.PyPI}`.
        """
        return {ECOSYSTEM.Npm, ECOSYSTEM.PyPI}

    def verify(self, package: Package) -> list[tuple[FindingSeverity, str]]:
        """
        Determine whether the given package is malicious by consulting the dataset's manifests.

        Args:
            package: The `Package` to verify.

        Returns:
            A list containing any findings for the given package, obtained by checking for its
            presence in the dataset's manifests.  Only a single `CRITICAL` finding to this effect
            is present in this case.
        """
        manifest = self._manifests.get(package.ecosystem)
        if not manifest:
            return [(FindingSeverity.WARNING, f"Package ecosystem {package.ecosystem} is not supported")]

        if (
            package.name in manifest
            and (not manifest[package.name] or package.version in manifest[package.name])
        ):
            return [
                (
                    FindingSeverity.CRITICAL,
                    f"Datadog Security Research has determined that package {package} is malicious"
                )
            ]
        else:
            return []


def load_verifier() -> PackageVerifier:
    """
    Export `DatadogMaliciousPackagesVerifier` for discovery by Supply-Chain Firewall.

    Returns:
        A `DatadogMaliciousPackagesVerifier` for use in a run of Supply-Chain Firewall.
    """
    return DatadogMaliciousPackagesVerifier()



"""
Provides utilities for configuring the environment (via `.rc` files) for using Supply-Chain Firewall.
"""

import logging
from pathlib import Path
import re

from scfw.constants import DD_AGENT_PORT_VAR, DD_API_KEY_VAR, DD_LOG_LEVEL_VAR, SCFW_HOME_VAR

_log = logging.getLogger(__name__)

_CONFIG_FILES = [".bashrc", ".zshrc"]

_BLOCK_START = "# BEGIN SCFW MANAGED BLOCK"
_BLOCK_END = "# END SCFW MANAGED BLOCK"


def update_config_files(answers: dict):
    """
    Update the Supply-Chain Firewall configuration in all supported files.

    Args:
        answers: A `dict` of configuration options to format and write to each file.
    """
    scfw_config = _format_answers(answers)

    for config_file in [Path.home() / file for file in _CONFIG_FILES]:
        if not config_file.exists():
            _log.info(f"Skipped adding configuration to file {config_file}: file does not already exist")
            continue

        _update_config_file(config_file, scfw_config)
        _log.info(f"Successfully added configuration to file {config_file}")


def _update_config_file(config_file: Path, scfw_config: str):
    """
    Update the Supply-Chain Firewall configuration in the given file.

    Args:
        config_file: A `Path` to the configuration file to update.
        scfw_config: A `str` containing the formatted configuration options to write.
    """
    scfw_block = f"{_BLOCK_START}\n{scfw_config}{_BLOCK_END}" if scfw_config else ""

    with open(config_file, "r+") as f:
        original_config = f.read()

        updated_config = re.sub(f"{_BLOCK_START}(.*?){_BLOCK_END}", scfw_block, original_config, flags=re.DOTALL)
        if updated_config == original_config and scfw_config not in original_config:
            updated_config = f"{original_config}\n{scfw_block}\n"

        f.seek(0)
        f.write(updated_config)
        f.truncate()


def _format_answers(answers: dict) -> str:
    """
    Format configuration options into .rc file `str` content.

    Args:
        answers: A `dict` containing the user's selected configuration options.

    Returns:
        A `str` containing the desired configuration content for writing into a .rc file.
    """
    config = ''

    if answers.get("alias_npm"):
        config += 'alias npm="scfw run npm"\n'
    if answers.get("alias_pip"):
        config += 'alias pip="scfw run pip"\n'
    if answers.get("alias_poetry"):
        config += 'alias poetry="scfw run poetry"\n'
    if (dd_agent_port := answers.get("dd_agent_port")):
        config += f'export {DD_AGENT_PORT_VAR}="{dd_agent_port}"\n'
    if (dd_api_key := answers.get("dd_api_key")):
        config += f'export {DD_API_KEY_VAR}="{dd_api_key}"\n'
    if (dd_log_level := answers.get("dd_log_level")):
        config += f'export {DD_LOG_LEVEL_VAR}="{dd_log_level}"\n'
    if (scfw_home := answers.get("scfw_home")):
        config += f'export {SCFW_HOME_VAR}="{scfw_home}"\n'

    return config



"""
A class for structuring and displaying the results of package verification.
"""

from collections.abc import Iterable
from typing import Optional
from typing_extensions import Self

from scfw.package import Package


class VerificationReport:
    """
    A structured report containing findings resulting from package verification.
    """
    def __init__(self) -> None:
        """
        Initialize a new, empty `VerificationReport`.
        """
        self._report: dict[Package, list[str]] = {}

    def __len__(self) -> int:
        """
        Return the number of entries in the report.
        """
        return len(self._report)

    def __str__(self) -> str:
        """
        Return a human-readable version of a verification report.

        Returns:
            A `str` containing the formatted verification report.
        """
        def show_line(linenum: int, line: str) -> str:
            return (f"  - {line}" if linenum == 0 else f"    {line}")

        def show_finding(finding: str) -> str:
            return '\n'.join(
                show_line(linenum, line) for linenum, line in enumerate(finding.split('\n'))
            )

        def show_findings(package: Package, findings: list[str]) -> str:
            return f"Package {package}:\n" + '\n'.join(map(show_finding, findings))

        return '\n'.join(
            show_findings(package, findings) for package, findings in self._report.items()
        )

    def get(self, package: Package) -> Optional[list[str]]:
        """
        Get the findings for the given package.

        Args:
            package: The `Package` to look up in the report.

        Returns:
            The reported findings for `package` or `None` if it is not present.
        """
        return self._report.get(package)

    def insert(self, package: Package, finding: str) -> None:
        """
        Insert the given package and finding into the report.

        Args:
            package: The `Package` to insert into the report.
            findings: The finding being reported for `package`.
        """
        if package in self._report:
            self._report[package].append(finding)
        else:
            self._report[package] = [finding]

    def extend(self, other: Self) -> None:
        """
        Extend a `VerificationReport` with additional findings from another.

        Args:
            other: The `VerificationReport` whose findings will be extended into `self`.
        """
        for package, findings in other._report.items():
            if package in self._report:
                self._report[package].extend(findings)
            else:
                self._report[package] = findings

    def packages(self) -> Iterable[Package]:
        """
        Return an iterator over `Package` mentioned in the report.
        """
        return (package for package in self._report)



"""
Exports the currently discoverable set of package verifiers for use in
Supply-Chain Firewall.

Two package verifiers ship with Supply-Chain Firewall by default: one for
Datadog Security Research's malicious packages dataset and one for OSV.dev's
advisory database. Users of Supply-Chain Firewall may additionally provide
custom verifiers representing alternative sources of truth.

Supply-Chain Firewall discovers verifiers at runtime via the following protocol.
The module implementing the custom verifier must contain a function with the
following name and signature:

```
def load_verifier() -> PackageVerifier
```

This `load_verifier` function should return an instance of the custom verifier.
The module may then be placed in the same directory as this source file for
runtime import. Make sure to reinstall Supply-Chain Firewall after doing so.
"""

import concurrent.futures as cf
import importlib
import itertools
import logging
import os
import pkgutil

from scfw.ecosystem import ECOSYSTEM
from scfw.package import Package
from scfw.report import VerificationReport
from scfw.verifier import FindingSeverity

_log = logging.getLogger(__name__)


class FirewallVerifiers:
    """
    Provides a simple interface to verifying packages against the set of currently
    discoverable verifiers.
    """
    def __init__(self, ecosystem: ECOSYSTEM):
        """
        Initialize a `FirewallVerifiers` from currently discoverable package verifiers
        that support the given package ecosystem.

        Raises:
            RuntimeError: No verifiers supporting the given package ecosystem currently discoverable.
        """
        self._verifiers = []

        for _, module, _ in pkgutil.iter_modules([os.path.dirname(__file__)]):
            try:
                verifier = importlib.import_module(f".{module}", package=__name__).load_verifier()
                if ecosystem in verifier.supported_ecosystems():
                    self._verifiers.append(verifier)
            except ModuleNotFoundError:
                _log.warning(f"Failed to load module {module} while collecting package verifiers")
            except AttributeError:
                _log.warning(f"Module {module} does not export a package verifier")

        if not self._verifiers:
            raise RuntimeError(f"No verifiers for package ecosystem {ecosystem} currently discoverable")

    def names(self) -> list[str]:
        """
        Return the names of discovered package verifiers.
        """
        return [verifier.name() for verifier in self._verifiers]

    def verify_packages(self, packages: list[Package]) -> dict[FindingSeverity, VerificationReport]:
        """
        Verify a set of packages against all discovered verifiers.

        Args:
            packages: The set of `Package` to verify.

        Returns:
            A set of severity-ranked verification reports resulting from verifying
            `packages` against all discovered verifiers.
        """
        reports: dict[FindingSeverity, VerificationReport] = {}

        with cf.ThreadPoolExecutor() as executor:
            task_results = {
                executor.submit(lambda v, t: v.verify(t), verifier, package): (verifier.name(), package)
                for verifier, package in itertools.product(self._verifiers, packages)
            }
            for future in cf.as_completed(task_results):
                verifier, package = task_results[future]
                if (findings := future.result()):
                    _log.info(f"Verifier {verifier} had findings for package {package}")
                    for severity, finding in findings:
                        if severity not in reports:
                            reports[severity] = VerificationReport()
                        reports[severity].insert(package, finding)
                else:
                    _log.info(f"Verifier {verifier} had no findings for package {package}")

        _log.info("Verification of packages complete")
        return reports



"""
Exposes the currently discoverable set of client loggers implementing the
firewall's logging protocol.

Two loggers ship with the supply chain firewall by default: `DDAgentLogger`
and `DDAPILogger`, which send logs to Datadog via a local Datadog Agent
or the HTTP API, respectively. Firewall users may additionally provide custom
loggers according to their own logging needs.

The firewall discovers loggers at runtime via the following simple protocol.
The module implementing the custom logger must contain a function with the
following name and signature:

```
def load_logger() -> FirewallLogger
```

This `load_logger` function should return an instance of the custom logger
for the firewall's use. The module may then be placed in the same directory
as this source file for runtime import. Make sure to reinstall the package
after doing so.
"""

import importlib
import logging
import os
import pkgutil

from scfw.ecosystem import ECOSYSTEM
from scfw.logger import FirewallAction, FirewallLogger
from scfw.package import Package
from scfw.report import VerificationReport
from scfw.verifier import FindingSeverity

_log = logging.getLogger(__name__)


class FirewallLoggers(FirewallLogger):
    """
    A `FirewallLogger` that logs to all currently discoverable `FirewallLoggers`.
    """
    def __init__(self):
        """
        Initialize a new `FirewallLoggers` instance from currently discoverable loggers.
        """
        self._loggers = []

        for _, module, _ in pkgutil.iter_modules([os.path.dirname(__file__)]):
            try:
                logger = importlib.import_module(f".{module}", package=__name__).load_logger()
                self._loggers.append(logger)
            except ModuleNotFoundError:
                _log.warning(f"Failed to load module {module} while collecting loggers")
            except AttributeError:
                _log.debug(f"Module {module} does not export a logger")
            except Exception as e:
                _log.warning(f"Failed to initialize logger defined in {module}: {e}")

        if not self._loggers:
            _log.warning("No loggers were discovered and successfully initialized")

    def log_firewall_action(
        self,
        ecosystem: ECOSYSTEM,
        package_manager: str,
        executable: str,
        command: list[str],
        targets: list[Package],
        action: FirewallAction,
        verified: bool,
        warned: bool,
    ):
        """
        Log the data and action taken in a completed run of Supply-Chain Firewall to
        all client loggers.
        """
        for logger in self._loggers:
            try:
                logger.log_firewall_action(
                    ecosystem,
                    package_manager,
                    executable,
                    command,
                    targets,
                    action,
                    verified,
                    warned,
                )
            except Exception as e:
                _log.warning(f"Failed to log firewall action: {e}")

    def log_audit(
        self,
        ecosystem: ECOSYSTEM,
        package_manager: str,
        executable: str,
        reports: dict[FindingSeverity, VerificationReport],
    ):
        """
        Log the results of an audit to all client loggers.
        """
        for logger in self._loggers:
            try:
                logger.log_audit(ecosystem, package_manager, executable, reports)
            except Exception as e:
                _log.warning(f"Failed to log audit: {e}")



"""
Provides a base class for package verifiers.
"""

from abc import (ABCMeta, abstractmethod)
from enum import Enum

from scfw.ecosystem import ECOSYSTEM
from scfw.package import Package


class FindingSeverity(Enum):
    """
    A hierarchy of severity levels for package verifier findings.

    Package verifiers attach severity levels to their findings in order to direct
    Supply-Chain Firewall to take the correct action with respect to blocking or
    warning on a package manager command.

    A `CRITICAL` finding causes Supply-Chain Firewall to block. A `WARNING` finding
    prompts it to seek confirmation from the user before running the command.
    """
    CRITICAL = "CRITICAL"
    WARNING = "WARNING"

    def __str__(self) -> str:
        """
        Format a `FindingSeverity` for printing.

        Returns:
            A `str` representing the given `FindingSeverity` suitable for printing.
        """
        return self.name


class PackageVerifier(metaclass=ABCMeta):
    """
    Abstract base class for package verifiers.

    Each package verifier should implement a service for verifying packages in all
    supported ecosystems against a single reputable source of data on vulnerable and
    malicious open source packages.
    """
    @classmethod
    @abstractmethod
    def name(cls) -> str:
        """
        Return the verifier's name.

        Returns:
            A constant, short, descriptive name `str` identifying the verifier.
        """
        pass

    @classmethod
    @abstractmethod
    def supported_ecosystems(cls) -> set[ECOSYSTEM]:
        """
        Return the set of package ecosystems the verifier supports.

        Returns:
            A constant `set` of `ECOSYSTEM` representing the package ecosystems
            supported for verification by the verifier.
        """
        pass

    @abstractmethod
    def verify(self, package: Package) -> list[tuple[FindingSeverity, str]]:
        """
        Verify the given package.

        Args:
            package: The `Package` to verify.

        Returns:
            A `list[tuple[FindingSeverity, str]]` of all findings for the given package
            reported by the backing data source, each tagged with a severity level.

            Each `str` in this list should be a concise summary of a single finding and
            would ideally provide a link or handle to more information about that finding
            for the benefit of the user.
        """
        pass



"""
Tests of the Supply-Chain Firewall command-line interface.
"""

import pytest

from scfw.cli import _parse_command_line, _DEFAULT_LOG_LEVEL, Subcommand


def test_cli_no_options_no_command():
    """
    Invocation with no options or arguments.
    """
    argv = ["scfw"]
    args, _ = _parse_command_line(argv)
    assert args is None


def test_cli_all_options_no_command():
    """
    Invocation with all top-level options and no subcommand.
    """
    argv = ["scfw", "--log-level", "DEBUG"]
    args, _ = _parse_command_line(argv)
    assert args is None


def test_cli_incorrect_subcommand():
    """
    Invocation with a nonexistent subcommand.
    """
    argv = ["scfw", "nonexistent"]
    args, _ = _parse_command_line(argv)
    assert args is None


@pytest.mark.parametrize(
        "target",
        ["npm", "pip", "poetry"]
)
def test_cli_audit_basic_usage(target: str):
    """
    Test of basic audit command usage for the given package manager `target`.
    """
    argv = ["scfw", "audit", target]
    args, _ = _parse_command_line(argv)
    assert args is not None

    assert args.subcommand == Subcommand.Audit
    assert args.log_level == _DEFAULT_LOG_LEVEL

    assert args.package_manager == target
    assert args.executable is None


def test_cli_audit_no_options_no_command():
    """
    Invocation of the audit subcommand with no options or arguments.
    """
    argv = ["scfw", "audit"]
    args, _ = _parse_command_line(argv)
    assert args is None


def test_cli_audit_all_options_no_package_manager():
    """
    Invocation of the audit subcommand with all options and no package manager.
    """
    executable = "/path/to/executable"
    argv = ["scfw", "audit", "--executable", executable]
    args, _ = _parse_command_line(argv)
    assert args is None


def test_cli_audit_unknown_package_manager():
    """
    Invocation of the audit subcommand with an unknown package manager.
    """
    argv = ["scfw", "audit", "foo"]
    args, _ = _parse_command_line(argv)
    assert args is None


@pytest.mark.parametrize(
        "target",
        ["npm", "pip", "poetry"]
)
def test_cli_audit_all_options_package_manager(target: str):
    """
    Invocation of an audit command with all options and the given package manager `target`.
    """
    executable = "/path/to/executable"
    argv = ["scfw", "audit", "--executable", executable, target]
    args, _ = _parse_command_line(argv)
    assert args is not None

    assert args.subcommand == Subcommand.Audit
    assert args.log_level == _DEFAULT_LOG_LEVEL

    assert args.package_manager == target
    assert args.executable == executable


def test_cli_configure_basic_usage():
    """
    Basic `configure` subcommand usage.
    """
    argv = ["scfw", "configure"]
    args, _ = _parse_command_line(argv)
    assert args is not None

    assert args.subcommand == Subcommand.Configure
    assert args.log_level == _DEFAULT_LOG_LEVEL

    assert not args.remove
    assert not args.alias_npm
    assert not args.alias_pip
    assert not args.alias_poetry
    assert args.dd_agent_port is None
    assert args.dd_api_key is None
    assert args.dd_log_level is None
    assert args.scfw_home is None


@pytest.mark.parametrize(
        "option",
        [
            ["--alias-npm"],
            ["--alias-pip"],
            ["--alias-poetry"],
            ["--dd-agent-port", "10365"],
            ["--dd-api-key", "foo"],
            ["--dd-log-level", "BLOCK"],
            ["--scfw-home", "foo"],
        ]
)
def test_cli_configure_removal(option: list[str]):
    """
    Test that the `--remove` configure option is not allowed with `option`.
    """
    argv = ["scfw", "configure", "--remove"] + option
    args, _ = _parse_command_line(argv)
    assert args is None


@pytest.mark.parametrize(
        "command",
        [
            ["npm", "install", "react"],
            ["pip", "install", "requests"],
            ["poetry", "add", "requests"],
        ]
)
def test_cli_run_basic_usage(command: list[str]):
    """
    Test of basic run command usage for the given package manager `command`.
    """
    argv = ["scfw", "run"] + command
    args, _ = _parse_command_line(argv)
    assert args is not None

    assert args.subcommand == Subcommand.Run
    assert args.log_level == _DEFAULT_LOG_LEVEL

    assert args.package_manager == argv[2]
    assert args.command == argv[2:]
    assert not args.allow_unsupported
    assert not args.dry_run
    assert not args.executable


def test_cli_run_all_options_no_command():
    """
    Invocation with all options and no arguments.
    """
    executable = "/usr/bin/python"
    argv = ["scfw", "run", "--executable", executable, "--dry-run", "--allow-unsupported"]
    args, _ = _parse_command_line(argv)
    assert args is None


@pytest.mark.parametrize(
        "command",
        [
            ["npm", "install", "react"],
            ["pip", "install", "requests"],
            ["poetry", "add", "requests"],
        ]
)
def test_cli_run_all_options_command(command: list[str]):
    """
    Invocation of a run command with all options and the given `command`.
    """
    executable = "/path/to/executable"
    argv = ["scfw", "run", "--executable", executable, "--dry-run", "--allow-unsupported"] + command
    args, _ = _parse_command_line(argv)
    assert args is not None

    assert args.subcommand == Subcommand.Run
    assert args.log_level == _DEFAULT_LOG_LEVEL

    assert args.package_manager == argv[6]
    assert args.command == argv[6:]
    assert args.allow_unsupported
    assert args.dry_run
    assert args.executable == executable


@pytest.mark.parametrize(
        "command",
        [
            ["npm", "install", "react"],
            ["pip", "install", "requests"],
            ["poetry", "install", "requests"],
        ]
)
def test_cli_run_package_manager_dry_run(command: list[str]):
    """
    Test that a `--dry-run` flag belonging to the package manager command
    is parsed correctly as such.
    """
    argv = ["scfw", "run"] + command + ["--dry-run"]
    args, _ = _parse_command_line(argv)
    assert args is not None

    assert args.subcommand == Subcommand.Run
    assert args.log_level == _DEFAULT_LOG_LEVEL

    assert args.package_manager == argv[2]
    assert args.command == argv[2:]
    assert not args.dry_run
    assert not args.executable


@pytest.mark.parametrize(
        "target,test",
        [
            ("npm", "pip"),
            ("npm", "poetry"),
            ("pip", "npm"),
            ("pip", "poetry"),
            ("poetry", "npm"),
            ("poetry", "pip"),
        ]
)
def test_cli_run_priority(target: str, test: str):
    """
    Test that a `target` command is parsed correctly in the presence of a `test` literal.
    """
    argv = ["scfw", "run", target, "foo", test]
    args, _ = _parse_command_line(argv)
    assert args is not None

    assert args.package_manager == argv[2]
    assert args.command == argv[2:]



"""
Provides a `PackageManager` representation of `npm`.
"""

import json
import logging
import os
import shutil
import subprocess
from typing import Optional

from packaging.version import InvalidVersion, Version, parse as version_parse

from scfw.ecosystem import ECOSYSTEM
from scfw.package import Package
from scfw.package_manager import PackageManager, UnsupportedVersionError

_log = logging.getLogger(__name__)

MIN_NPM_VERSION = version_parse("7.0.0")


class Npm(PackageManager):
    """
    A `PackageManager` representation of `npm`.
    """
    def __init__(self, executable: Optional[str] = None):
        """
        Initialize a new `Npm` instance.

        Args:
            executable:
                An optional path in the local filesystem to the `npm` executable to use.
                If not provided, this value is determined by the current environment.

        Raises:
            RuntimeError: A valid executable could not be resolved.
        """
        executable = executable if executable else shutil.which(self.name())
        if not executable:
            raise RuntimeError("Failed to resolve local npm executable: is npm installed?")
        if not os.path.isfile(executable):
            raise RuntimeError(f"Path '{executable}' does not correspond to a regular file")

        self._executable = executable

    @classmethod
    def name(cls) -> str:
        """
        Return the token for invoking `npm` on the command line.
        """
        return "npm"

    @classmethod
    def ecosystem(cls) -> ECOSYSTEM:
        """
        Return the ecosystem of packages managed by `npm`.
        """
        return ECOSYSTEM.Npm

    def executable(self) -> str:
        """
        Return the local filesystem path to the underlying `npm` executable.
        """
        return self._executable

    def run_command(self, command: list[str]) -> int:
        """
        Run an `npm` command.

        Args:
            command: A `list[str]` containing an `npm` command to execute.

        Returns:
            An `int` return code describing the exit status of the executed `npm` command.

        Raises:
            ValueError: The given `command` is empty or not a valid `npm` command.
        """
        return subprocess.run(self._normalize_command(command)).returncode

    def resolve_install_targets(self, command: list[str]) -> list[Package]:
        """
        Resolve the installation targets of the given `npm` command.

        Args:
            command:
                A `list[str]` representing an `npm` command whose installation targets
                are to be resolved.

        Returns:
            A `list[Package]` representing the package targets that would be installed
            if `command` were run.

        Raises:
            ValueError:
                1) The given `command` is empty or not a valid `npm` command, or 2) failed to parse
                an installation target.
            UnsupportedVersionError: The underlying `npm` executable is of an unsupported version.
        """
        def is_install_command(command: list[str]) -> bool:
            # https://docs.npmjs.com/cli/v10/commands/npm-install
            install_aliases = {
                "install", "add", "i", "in", "ins", "inst", "insta", "instal", "isnt", "isnta", "isntal", "isntall"
            }
            return any(alias in command for alias in install_aliases)

        def is_place_dep_line(line: str) -> bool:
            # The "placeDep" log lines describe a new dependency added to the
            # dependency tree being constructed by an installish command
            return "placeDep" in line

        def line_to_dependency(line: str) -> str:
            # Each added dependency is always the fifth token in its log line
            return line.split()[4]

        def str_to_package(s: str) -> Package:
            name, sep, version = s.rpartition('@')
            if version == s or (sep and not name):
                raise ValueError("Failed to parse npm installation target")
            return Package(ECOSYSTEM.Npm, name, version)

        command = self._normalize_command(command)

        # For now, allow all non-`install` commands
        if not is_install_command(command):
            return []

        self._check_version()

        # On supported versions, the presence of these options prevents the command from running
        if any(opt in command for opt in {"-h", "--help", "--dry-run"}):
            return []

        try:
            # Compute the set of dependencies added by the install command
            dry_run_command = command + ["--dry-run", "--loglevel", "silly"]
            dry_run = subprocess.run(dry_run_command, check=True, text=True, capture_output=True)
            dependencies = map(line_to_dependency, filter(is_place_dep_line, dry_run.stderr.strip().split('\n')))
        except subprocess.CalledProcessError:
            # An erroring command does not install anything
            _log.info("Encountered an error while resolving npm installation targets")
            return []

        try:
            # List targets already installed in the npm environment
            list_command = [self.executable(), "list", "--all"]
            installed = subprocess.run(list_command, check=True, text=True, capture_output=True).stdout
        except subprocess.CalledProcessError:
            # If this operation fails, rather than blocking, assume nothing is installed
            # This has the effect of treating all dependencies like installation targets
            _log.warning(
                "Failed to list installed npm packages: treating all dependencies as installation targets"
            )
            installed = ""

        # The installation targets are the dependencies that are not already installed
        targets = filter(lambda dep: dep not in installed, dependencies)

        return list(map(str_to_package, targets))

    def list_installed_packages(self) -> list[Package]:
        """
        List all `npm` packages installed in the active `npm` environment.

        Returns:
            A `list[Package]` representing all `npm` packages installed in the active
            `npm` environment.

        Raises:
            RuntimeError: Failed to list installed packages or decode report JSON.
            ValueError: Encountered a malformed report for an installed package.
            UnsupportedVersionError: The underlying `npm` executable is of an unsupported version.
        """
        def dependencies_to_packages(dependencies: dict[str, dict]) -> set[Package]:
            packages = set()

            for name, package_data in dependencies.items():
                if (package_dependencies := package_data.get("dependencies")):
                    packages |= dependencies_to_packages(package_dependencies)
                packages.add(Package(ECOSYSTEM.Npm, name, package_data["version"]))

            return packages

        self._check_version()

        try:
            npm_list_command = self._normalize_command(["npm", "list", "--all", "--json"])
            npm_list = subprocess.run(npm_list_command, check=True, text=True, capture_output=True)
            dependencies = json.loads(npm_list.stdout.strip()).get("dependencies")
            return list(dependencies_to_packages(dependencies)) if dependencies else []

        except subprocess.CalledProcessError:
            raise RuntimeError("Failed to list npm installed packages")

        except json.JSONDecodeError:
            raise RuntimeError("Failed to decode installed package report JSON")

        except KeyError:
            raise ValueError("Malformed installed package report")

    def _check_version(self):
        """
        Check whether the underlying `npm` executable is of a supported version.

        Raises:
            UnsupportedVersionError: The underlying `npm` executable is of an unsupported version.
        """
        def get_npm_version(executable: str) -> Optional[Version]:
            try:
                # All supported versions adhere to this format
                npm_version = subprocess.run([executable, "--version"], check=True, text=True, capture_output=True)
                return version_parse(npm_version.stdout.strip())
            except InvalidVersion:
                return None

        npm_version = get_npm_version(self._executable)
        if not npm_version or npm_version < MIN_NPM_VERSION:
            raise UnsupportedVersionError(f"npm before v{MIN_NPM_VERSION} is not supported")

    def _normalize_command(self, command: list[str]) -> list[str]:
        """
        Normalize an `npm` command.

        Args:
            command: A `list[str]` containing an `npm` command line.

        Returns:
            The equivalent but normalized form of `command` with the initial `npm`
            token replaced with the local filesystem path to `self.executable()`.

        Raises:
            ValueError: The given `command` is empty or not a valid `npm` command.
        """
        if not command:
            raise ValueError("Received empty npm command line")
        if command[0] != self.name():
            raise ValueError("Received invalid npm command line")

        return [self._executable] + command[1:]



"""
Provides a `PackageManager` representation of `poetry`.
"""

import logging
import os
import re
import shutil
import subprocess
from typing import Optional

from packaging.version import InvalidVersion, Version, parse as version_parse

from scfw.ecosystem import ECOSYSTEM
from scfw.package import Package
from scfw.package_manager import PackageManager, UnsupportedVersionError

_log = logging.getLogger(__name__)

MIN_POETRY_VERSION = version_parse("1.7.0")

INSPECTED_SUBCOMMANDS = {"add", "install", "sync", "update"}


class Poetry(PackageManager):
    """
    A `PackageManager` representation of `poetry`.
    """
    def __init__(self, executable: Optional[str] = None):
        """
        Initialize a new `Poetry` instance.

        Args:
            executable:
                An optional path in the local filesystem to the `poetry` executable to use.
                If not provided, this value is determined by the current environment.

        Raises:
            RuntimeError: A valid executable could not be resolved.
        """
        executable = executable if executable else shutil.which(self.name())
        if not executable:
            raise RuntimeError("Failed to resolve local Poetry executable: is Poetry installed?")
        if not os.path.isfile(executable):
            raise RuntimeError(f"Path '{executable}' does not correspond to a regular file")

        self._executable = executable

    @classmethod
    def name(cls) -> str:
        """
        Return the token for invoking `poetry` on the command line.
        """
        return "poetry"

    @classmethod
    def ecosystem(cls) -> ECOSYSTEM:
        """
        Return the ecosystem of packages managed by `poetry`.
        """
        return ECOSYSTEM.PyPI

    def executable(self) -> str:
        """
        Return the local filesystem path to the underlying `poetry` executable.
        """
        return self._executable

    def run_command(self, command: list[str]) -> int:
        """
        Run a `poetry` command.

        Args:
            command: A `list[str]` containing a `poetry` command to execute.

        Returns:
            An `int` return code describing the exit status of the executed `poetry` command.

        Raises:
            ValueError: The given `command` is empty or not a valid `poetry` command.
        """
        return subprocess.run(self._normalize_command(command)).returncode

    def resolve_install_targets(self, command: list[str]) -> list[Package]:
        """
        Resolve the installation targets of the given `poetry` command.

        Args:
            command:
                A `list[str]` representing a `poetry` command whose installation targets
                are to be resolved.

        Returns:
            A `list[Package]` representing the package targets that would be installed
            if `command` were run.

        Raises:
            ValueError: The given `command` is empty or not a valid `poetry` command.
            UnsupportedVersionError: The underlying `poetry` executable is of an unsupported version.
        """
        def get_target_version(version_spec: str) -> str:
            _, arrow, new_version = version_spec.partition(" -> ")
            version, _, _ = version_spec.partition(' ')
            return get_target_version(new_version) if arrow else version

        def line_to_package(line: str) -> Optional[Package]:
            # All supported versions adhere to this format
            pattern = r"(Installing|Updating|Downgrading) (?:the current project: )?(.*) \((.*)\)"
            if "Skipped" not in line and (match := re.search(pattern, line.strip())):
                return Package(self.ecosystem(), match.group(2), get_target_version(match.group(3)))
            return None

        command = self._normalize_command(command)

        if not any(subcommand in command for subcommand in INSPECTED_SUBCOMMANDS):
            return []

        self._check_version()

        # On supported versions, the presence of these options prevents the command from running
        if any(opt in command for opt in {"-V", "--version", "-h", "--help", "--dry-run"}):
            return []

        try:
            # Compute installation targets: new dependencies and updates/downgrades of existing ones
            dry_run = subprocess.run(command + ["--dry-run"], check=True, text=True, capture_output=True)
            return list(filter(None, map(line_to_package, dry_run.stdout.split('\n'))))
        except subprocess.CalledProcessError:
            # An erroring command does not install anything
            _log.info("Encountered an error while resolving poetry installation targets")
            return []

    def list_installed_packages(self) -> list[Package]:
        """
        List all `PyPI` packages installed in the active `poetry` environment.

        Returns:
            A `list[Package]` representing all `PyPI` packages installed in the active
            `poetry` environment.

        Raises:
            RuntimeError: Failed to list installed packages.
            ValueError: Malformed installed package report.
            UnsupportedVersionError: The underlying `poetry` executable is of an unsupported version.
        """
        def line_to_package(line: str) -> Package:
            tokens = line.split()
            return Package(ECOSYSTEM.PyPI, tokens[0], tokens[1])

        self._check_version()

        try:
            poetry_show_command = self._normalize_command(["poetry", "show", "--all"])
            poetry_show = subprocess.run(poetry_show_command, check=True, text=True, capture_output=True)
            installed_report = poetry_show.stdout.strip()
            return list(map(line_to_package, installed_report.split('\n'))) if installed_report else []

        except subprocess.CalledProcessError:
            raise RuntimeError("Failed to list poetry installed packages")

        except IndexError:
            raise ValueError("Malformed installed package report")

    def _check_version(self):
        """
        Check whether the underlying `poetry` executable is of a supported version.

        Raises:
            UnsupportedVersionError: The underlying `poetry` executable is of an unsupported version.
        """
        def get_poetry_version(executable: str) -> Optional[Version]:
            try:
                # All supported versions adhere to this format
                poetry_version = subprocess.run([executable, "--version"], check=True, text=True, capture_output=True)
                match = re.search(r"Poetry \(version (.*)\)", poetry_version.stdout.strip())
                return version_parse(match.group(1)) if match else None
            except InvalidVersion:
                return None

        poetry_version = get_poetry_version(self._executable)
        if not poetry_version or poetry_version < MIN_POETRY_VERSION:
            raise UnsupportedVersionError(f"Poetry before v{MIN_POETRY_VERSION} is not supported")

    def _normalize_command(self, command: list[str]) -> list[str]:
        """
        Normalize a `poetry` command.

        Args:
            command: A `list[str]` containing a `poetry` command line.

        Returns:
            The equivalent but normalized form of `command` with the initial `poetry`
            token replaced with the local filesystem path to `self.executable()`.

        Raises:
            ValueError: The given `command` is empty or not a valid `poetry` command.
        """
        if not command:
            raise ValueError("Received empty poetry command line")
        if command[0] != self.name():
            raise ValueError("Received invalid poetry command line")

        return [self._executable] + command[1:]



"""
Defines a package verifier for the OSV.dev advisory database.
"""

import functools
import logging

import requests

from scfw.ecosystem import ECOSYSTEM
from scfw.package import Package
from scfw.verifier import FindingSeverity, PackageVerifier
from scfw.verifiers.osv_verifier.osv_advisory import OsvAdvisory

_log = logging.getLogger(__name__)

_OSV_DEV_QUERY_URL = "https://api.osv.dev/v1/query"
_OSV_DEV_VULN_URL_PREFIX = "https://osv.dev/vulnerability"
_OSV_DEV_LIST_URL_PREFIX = "https://osv.dev/list"


class OsvVerifier(PackageVerifier):
    """
    A `PackageVerifier` for the OSV.dev advisory database.
    """
    @classmethod
    def name(cls) -> str:
        """
        Return the `OsvVerifier` name string.

        Returns:
            The class' constant name string: `"OsvVerifier"`.
        """
        return "OsvVerifier"

    @classmethod
    def supported_ecosystems(cls) -> set[ECOSYSTEM]:
        """
        Return the set of package ecosystems supported by `OsvVerifier`.

        Returns:
            The class' constant set of supported ecosystems: `{ECOSYSTEM.Npm, ECOSYSTEM.PyPI}`.
        """
        return {ECOSYSTEM.Npm, ECOSYSTEM.PyPI}

    def verify(self, package: Package) -> list[tuple[FindingSeverity, str]]:
        """
        Query a given package against the OSV.dev database.

        Args:
            package: The `Package` to query.

        Returns:
            A list containing any findings for the given package, obtained by querying
            the OSV.dev API.

            OSV.dev advisories with `MAL` IDs are treated as `CRITICAL` findings and all
            others are treated as `WARNING`.  *It is very important to note that most but
            **not all** OSV.dev malicious package advisories have `MAL` IDs.*

        Raises:
            requests.HTTPError:
                An error occurred while querying a package against the OSV.dev API.
        """
        def finding(osv: OsvAdvisory) -> str:
            kind = "malicious package " if osv.id.startswith("MAL") else ""
            severity_tag = f"[{osv.severity}] " if osv.severity else ""
            return (
                f"An OSV.dev {kind}advisory exists for package {package}:\n"
                f"  * {severity_tag}{_OSV_DEV_VULN_URL_PREFIX}/{osv.id}"
            )

        def failure_message() -> str:
            return (
                f"Failed to verify package {package} via the OSV.dev API.\n"
                f"Before proceeding, please check the OSV.dev website for advisories related to this package.\n"
                f"DO NOT PROCEED if the package has advisories with a MAL ID: it is very likely malicious.\n"
                f"  * {_OSV_DEV_LIST_URL_PREFIX}?q={package.name}&ecosystem={str(package.ecosystem)}"
            )

        if package.ecosystem not in self.supported_ecosystems():
            return [(FindingSeverity.WARNING, f"Package ecosystem {package.ecosystem} is not supported")]

        vulns = []
        query = {
            "version": package.version,
            "package": {
                "name": package.name,
                "ecosystem": str(package.ecosystem)
            }
        }

        try:
            while True:
                # The OSV.dev API is sometimes quite slow, hence the generous timeout
                request = requests.post(_OSV_DEV_QUERY_URL, json=query, timeout=10)
                request.raise_for_status()
                response = request.json()

                if (response_vulns := response.get("vulns")):
                    vulns.extend(response_vulns)

                query["page_token"] = response.get("next_page_token")

                if not query["page_token"]:
                    break

            if not vulns:
                return []

            osvs = set(map(OsvAdvisory.from_json, filter(lambda vuln: vuln.get("id"), vulns)))
            mal_osvs = set(filter(lambda osv: osv.id.startswith("MAL"), osvs))
            non_mal_osvs = osvs - mal_osvs

            osv_sort_key = functools.cmp_to_key(OsvAdvisory.compare_severities)
            sorted_mal_osvs = sorted(mal_osvs, reverse=True, key=osv_sort_key)
            sorted_non_mal_osvs = sorted(non_mal_osvs, reverse=True, key=osv_sort_key)

            return (
                [(FindingSeverity.CRITICAL, finding(osv)) for osv in sorted_mal_osvs]
                + [(FindingSeverity.WARNING, finding(osv)) for osv in sorted_non_mal_osvs]
            )

        except requests.exceptions.RequestException as e:
            _log.warning(f"Failed to query OSV.dev API for package {package}: {e}")
            return [(FindingSeverity.WARNING, failure_message())]

        except Exception as e:
            _log.warning(f"Verification failed for package {package}: {e}")
            return [(FindingSeverity.WARNING, failure_message())]


def load_verifier() -> PackageVerifier:
    """
    Export `OsvVerifier` for discovery by Supply-Chain Firewall.

    Returns:
        An `OsvVerifier` for use in a run of Supply-Chain Firewall.
    """
    return OsvVerifier()



"""
Provides a base class for representing supported package managers.
"""

from abc import (ABCMeta, abstractmethod)
from typing import Optional

from scfw.ecosystem import ECOSYSTEM
from scfw.package import Package


class PackageManager(metaclass=ABCMeta):
    """
    Abstract base class for representing supported package managers.
    """
    @abstractmethod
    def __init__(self, executable: Optional[str] = None):
        """
        Initialize a new `PackageManager`.

        Args:
            executable:
                An optional local filesystem path to the underlying package manager executable
                that should be used for running commands. If none is provided, the executable
                is determined by the current environment.

        Raises:
            UnsupportedVersionError:
                Implementors should raise this error when the underlying executable has an
                unsupported version.
        """
        pass

    @classmethod
    @abstractmethod
    def name(cls) -> str:
        """
        Return the name of the package manager, the standard fixed token by which
        it is invoked on the command line.
        """
        pass

    @classmethod
    @abstractmethod
    def ecosystem(cls) -> ECOSYSTEM:
        """
        Return the fixed package ecosystem the package manager is for.
        """
        pass

    @abstractmethod
    def executable(self) -> str:
        """
        Return the local filesystem path to the package manager executable.
        """
        pass

    @abstractmethod
    def run_command(self, command: list[str]) -> int:
        """
        Run the given package manager command.

        Args:
            command: The package manager command to be run.

        Returns:
            An `int` return code describing the exit status of the executed command.
        """
        pass

    @abstractmethod
    def resolve_install_targets(self, command: list[str]) -> list[Package]:
        """
        Resolve the package targets that would be installed if the given package
        manager command were run (without running it).

        Args:
            command: The package manager command whose installation targets are to be resolved.

        Returns:
            A `list[Package]` representing the package targets that would be installed
            if `command` were run.
        """
        pass

    @abstractmethod
    def list_installed_packages(self) -> list[Package]:
        """
        List all installed packages.

        Returns:
            A `list[Package]` representing all currently installed packages.
        """
        pass


class UnsupportedVersionError(Exception):
    """
    An exception that occurs when an attempt is made to initialize a `PackageManager`
    with an unsupported version of the underlying executable. Supply-Chain Firewall
    handles this exception gracefully by alerting the user to the issue. In firewall
    mode, the user can optionally disable verification and run commands with
    unsupported versions of supported package managers.
    """
    pass



"""
Implements Supply-Chain Firewall's `audit` subcommand.
"""

from argparse import Namespace
import logging

from scfw.loggers import FirewallLoggers
import scfw.package_managers as package_managers
from scfw.report import VerificationReport
from scfw.verifier import FindingSeverity
from scfw.verifiers import FirewallVerifiers

_log = logging.getLogger(__name__)


def run_audit(args: Namespace) -> int:
    """
    Audit installed packages using Supply-Chain Firewall's verifiers.

    Args:
        args: A `Namespace` containing the parsed `audit` subcommand command line.

    Returns:
        An integer status code indicating normal exit.
    """
    merged_report = VerificationReport()

    package_manager = package_managers.get_package_manager(args.package_manager, executable=args.executable)

    if (packages := package_manager.list_installed_packages()):
        _log.info(f"Installed packages: [{', '.join(map(str, packages))}]")

        verifiers = FirewallVerifiers(package_manager.ecosystem())
        _log.info(f"Using package verifiers: [{', '.join(verifiers.names())}]")

        reports = verifiers.verify_packages(packages)
        FirewallLoggers().log_audit(
            package_manager.ecosystem(),
            package_manager.name(),
            package_manager.executable(),
            reports,
        )

        for severity in FindingSeverity:
            if (severity_report := reports.get(severity)):
                merged_report.extend(severity_report)

    if merged_report:
        print(merged_report)
    else:
        print("No issues found.")

    return 0



"""
Provides an interface for obtaining `PackageManager` instances needed for runs
of Supply-Chain Firewall.
"""

from typing import Optional

from scfw.package_manager import PackageManager
from scfw.package_managers.npm import Npm
from scfw.package_managers.pip import Pip
from scfw.package_managers.poetry import Poetry

SUPPORTED_PACKAGE_MANAGERS = [
    Npm.name(),
    Pip.name(),
    Poetry.name(),
]
"""
Contains the command line names of supported package managers.
"""


def get_package_manager(name: str, executable: Optional[str] = None) -> PackageManager:
    """
    Return a `PackageManager` corresponding to the given command line name.

    Args:
        name: The command line name of the desired `PackageManager`.
        executable: An optional executable to use to initialize the returned `PackageManager`.

    Returns:
        A `PackageManager` corresponding to `name` and initialized from `executable`.

    Raises:
        ValueError: An empty or unsupported package manager name was provided.
    """
    if not name:
        raise ValueError("Missing package manager")

    if name == Npm.name():
        return Npm(executable)
    if name == Pip.name():
        return Pip(executable)
    if name == Poetry.name():
        return Poetry(executable)

    raise ValueError(f"Unsupported package manager '{name}'")



"""
Configures a logger for sending firewall logs to a local Datadog Agent.
"""

import logging
import os
import socket

from scfw.constants import DD_AGENT_PORT_VAR
from scfw.logger import FirewallLogger
from scfw.loggers.dd_logger import DDLogFormatter, DDLogger

_log = logging.getLogger(__name__)

_DD_LOG_NAME = "dd_agent_log"


class _DDLogHandler(logging.Handler):
    def __init__(self, agent_port: int):
        """
        Initialize a new `_DDLogHandler` instance.

        Args:
            agent_port:
                The port number where the Datadog Agent has been independently configured
                to receive logs from Supply-Chain Firewall.

        Raises:
            ValueError: An invalid port number was given.
        """
        if not 0 < agent_port < 2 ** 16:
            raise ValueError(f"Invalid port number {agent_port}")
        self._agent_port = agent_port

        super().__init__()

    def emit(self, record):
        """
        Format and send a log to the Datadog Agent.

        Args:
            record: The log record to be forwarded.

        Raises:
            RuntimeError: Failed to forward log to Datadog Agent.
        """
        try:
            message = self.format(record).encode()

            s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            s.connect(("localhost", self._agent_port))
            if s.send(message) != len(message):
                raise ValueError("Failed to send log message over local socket")
            s.close()

        except Exception as e:
            raise RuntimeError(f"Failed to forward log to Datadog Agent: {e}")


# Configure a single logging handle for all `DDAgentLogger` instances to share
_handler: logging.Handler = logging.NullHandler()
if (dd_agent_port := os.getenv(DD_AGENT_PORT_VAR)):
    try:
        _handler = _DDLogHandler(int(dd_agent_port))
    except Exception as e:
        _log.warning(f"Failed to initialize Datadog Agent logger: {e}")
_handler.setFormatter(DDLogFormatter())

_ddlog = logging.getLogger(_DD_LOG_NAME)
_ddlog.setLevel(logging.INFO)
_ddlog.addHandler(_handler)


class DDAgentLogger(DDLogger):
    """
    An implementation of `FirewallLogger` for sending logs to a local Datadog Agent.
    """
    def __init__(self):
        """
        Initialize a new `DDAgentLogger`.
        """
        super().__init__(_ddlog)


def load_logger() -> FirewallLogger:
    """
    Export `DDAgentLogger` for discovery by the firewall.

    Returns:
        A `DDAgentLogger` for use in a run of the firewall.
    """
    return DDAgentLogger()



"""
Provides a `PackageManager` representation of `pip`.
"""

import json
import logging
import os
import shutil
import subprocess
from typing import Optional

from packaging.version import InvalidVersion, Version, parse as version_parse

from scfw.ecosystem import ECOSYSTEM
from scfw.package import Package
from scfw.package_manager import PackageManager, UnsupportedVersionError

_log = logging.getLogger(__name__)

MIN_PIP_VERSION = version_parse("22.2")


class Pip(PackageManager):
    """
    A `PackageManager` representation of `pip`.
    """
    def __init__(self, executable: Optional[str] = None):
        """
        Initialize a new `Pip` instance.

        Args:
            executable:
                An optional path in the local filesystem to the Python executable
                to use for running `pip` as a module. If not provided, this value
                is determined by the current environment.

        Raises:
            RuntimeError: A valid executable could not be resolved.
        """
        def get_python_executable() -> Optional[str]:
            # Explicitly checking whether we are in a venv circumvents issues
            # caused by pyenv shims stomping the PATH with its own directories
            venv_path = None
            if (venv := os.environ.get("VIRTUAL_ENV")):
                venv_path = os.path.join(venv, "bin")
            for bin in ["python3", "python"]:
                if (executable := shutil.which(bin, path=venv_path)):
                    return executable
            return None

        executable = executable if executable else get_python_executable()
        if not executable:
            raise RuntimeError("Failed to resolve local Python executable: is Python installed?")
        if not os.path.isfile(executable):
            raise RuntimeError(f"Path '{executable}' does not correspond to a regular file")

        self._executable = executable

    @classmethod
    def name(cls) -> str:
        """
        Return the token for invoking `pip` on the command line.
        """
        return "pip"

    @classmethod
    def ecosystem(cls) -> ECOSYSTEM:
        """
        Return the ecosystem of packages managed by `pip`.
        """
        return ECOSYSTEM.PyPI

    def executable(self) -> str:
        """
        Return the local filesystem path to the underlying `pip` executable.
        """
        return self._executable

    def run_command(self, command: list[str]) -> int:
        """
        Run a `pip` command.

        Args:
            command: A `list[str]` containing a `pip` command to execute.

        Returns:
            An `int` return code describing the exit status of the executed `pip` command.

        Raises:
            ValueError: The given `command` is empty or not a valid `pip` command.
        """
        return subprocess.run(self._normalize_command(command)).returncode

    def resolve_install_targets(self, command: list[str]) -> list[Package]:
        """
        Resolve the installation targets of the given `pip` command.

        Args:
            command:
                A `list[str]` representing a `pip` command whose installation targets
                are to be resolved.

        Returns:
            A `list[Package]` representing the package targets that would be installed
            if `command` were run.

        Raises:
            ValueError:
                1) The given `command` is empty or not a valid `pip` command, or 2) dry-run output
                did not have the required format.
            UnsupportedVersionError: The underlying `pip` executable is of an unsupported version.
        """
        def report_to_install_target(install_report: dict) -> Package:
            if not (metadata := install_report.get("metadata")):
                raise ValueError("Missing metadata for pip installation target")
            if not (name := metadata.get("name")):
                raise ValueError("Missing name for pip installation target")
            if not (version := metadata.get("version")):
                raise ValueError("Missing version for pip installation target")
            return Package(ECOSYSTEM.PyPI, name, version)

        command = self._normalize_command(command)

        # pip only installs or upgrades packages via the `pip install` subcommand
        # If `install` is not present, the command is automatically safe to run
        if "install" not in command:
            return []

        self._check_version()

        # On supported versions, the presence of these options prevents the command from running
        if any(opt in command for opt in {"-h", "--help", "--dry-run"}):
            return []

        # Otherwise, this is probably a live `pip install` command
        # To be certain, we would need to write a full parser for pip
        try:
            dry_run_command = command + ["--dry-run", "-qqqqq", "--report", "-"]
            dry_run = subprocess.run(dry_run_command, check=True, text=True, capture_output=True)
            install_reports = json.loads(dry_run.stdout).get("install", [])
            return list(map(report_to_install_target, install_reports))
        except subprocess.CalledProcessError:
            # An error must have resulted from the given pip command
            # As nothing will be installed in this case, allow the command
            _log.info("Encountered an error while resolving pip installation targets")
            return []

    def list_installed_packages(self) -> list[Package]:
        """
        List all `PyPI` packages installed in the active `pip` environment.

        Returns:
            A `list[Package]` representing all `PyPI` packages installed in the active
            `pip` environment.

        Raises:
            RuntimeError: Failed to list installed packages or decode report JSON.
            ValueError: Encountered a malformed report for an installed package.
            UnsupportedVersionError: The underlying `pip` executable is of an unsupported version.
        """
        self._check_version()

        try:
            pip_list_command = self._normalize_command(["pip", "list", "--format", "json"])
            pip_list = subprocess.run(pip_list_command, check=True, text=True, capture_output=True)
            return [
                Package(ECOSYSTEM.PyPI, package["name"], package["version"])
                for package in json.loads(pip_list.stdout.strip())
            ]

        except subprocess.CalledProcessError:
            raise RuntimeError("Failed to list pip installed packages")

        except json.JSONDecodeError:
            raise RuntimeError("Failed to decode installed package report JSON")

        except KeyError:
            raise ValueError("Malformed installed package report")

    def _check_version(self):
        """
        Check whether the underlying `pip` executable is of a supported version.

        Raises:
            UnsupportedVersionError: The underlying `pip` executable is of an unsupported version.
        """
        def get_pip_version(executable: str) -> Optional[Version]:
            try:
                pip_version = subprocess.run(
                    [executable, "-m", "pip", "--version"],
                    check=True,
                    text=True,
                    capture_output=True
                )
                # All supported versions adhere to this format
                version_str = pip_version.stdout.split()[1]
                return version_parse(version_str)
            except IndexError:
                return None
            except InvalidVersion:
                return None

        pip_version = get_pip_version(self._executable)
        if not pip_version or pip_version < MIN_PIP_VERSION:
            raise UnsupportedVersionError(f"pip before v{MIN_PIP_VERSION} is not supported")

    def _normalize_command(self, command: list[str]) -> list[str]:
        """
        Normalize a `pip` command.

        Args:
            command:
                A `list[str]` containing a "pure" `pip` command line (i.e., one
                that starts with `"pip"`)

        Returns:
            The equivalent but normalized form of `command` permitting Python
            module invocation of `pip`.

        Raises:
            ValueError: The given `command` is empty or not a valid `pip` command.
        """
        if not command:
            raise ValueError("Received empty pip command line")
        if command[0] != self.name():
            raise ValueError("Received invalid pip command line")

        return [self._executable, "-m"] + command
